#include "../stdafx.h"

#include "LineRenderer.h"
#include "Vertices.h"
#include "FrameTime.h"
#include "camera/Camera.h"
#include "../common/resource/TextureBundle.h"
#include "../common/resource/EffectBundle.h"
#include "../utils/DXHelpers.h"


CLineRenderer::CLineRenderer( CCamera const& camera, const CResourceId& effectId, const CResourceId& textureId, const TLinePool& lines )
:	mIB( NULL ), mVB( NULL ), mBufferPosition( 0 ),
	mCamera( camera ),
	mEffectId( effectId ),
	mTextureId( textureId ),
	mLines( lines )
{
}

CLineRenderer::~CLineRenderer()
{
}


// ------------------------------------------------------------------
//  DX part

// Buffers capacity in lines.
static const int BUFFERS_CAPACITY = 4000;


HRESULT CLineRenderer::onCreateDevice( IDirect3DDevice8* device )
{
	CDXObject::onCreateDevice( device );
	
	// IB
	assert( !mIB );
	getDevice().CreateIndexBuffer(
		BUFFERS_CAPACITY * 6 * sizeof(short),
		0,
		D3DFMT_INDEX16,
		D3DPOOL_MANAGED,
		&mIB
	);
	assert( mIB );

	short* ib = NULL;
	mIB->Lock( 0, 0, (BYTE**)&ib, 0 );
	assert( ib );
	for( int i = 0; i < BUFFERS_CAPACITY; ++i ) {
		short base = i * 4;
		ib[0] = base;
		ib[1] = base+1;
		ib[2] = base+2;
		ib[3] = base;
		ib[4] = base+2;
		ib[5] = base+3;
		ib += 6;
	}
	mIB->Unlock();
	
	return S_OK;
}

HRESULT CLineRenderer::onLostDevice()
{
	CDXObject::onLostDevice();
	dx::release( mVB );
	return S_OK;
}

HRESULT CLineRenderer::onResetDevice()
{
	CDXObject::onResetDevice();
	
	// VB
	assert( !mVB );
	getDevice().CreateVertexBuffer(
		BUFFERS_CAPACITY * 4 * sizeof(SVertexXyzDiffuseTex1),
		D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY,
		FVF_XYZ_DIFFUSE_TEX1,
		D3DPOOL_DEFAULT,
		&mVB
	);
	assert( mVB );

	mBufferPosition = 0;
	
	return S_OK;
}

HRESULT CLineRenderer::onDestroyDevice()
{
	CDXObject::onDestroyDevice();
	
	dx::release( mIB );
	
	return S_OK;
}


// ------------------------------------------------------------------
//  Rendering

void CLineRenderer::render( CFrameTime const& frameTime )
{
	if( !isVisible() )
		return;

	if( mLines.empty() )
		return;

	D3DXMATRIX viewMatrix = mCamera.getCameraMatrix();
	viewMatrix._41 = viewMatrix._42 = viewMatrix._43 = 0.0f;
	const D3DXVECTOR3& camPos = mCamera.getPosition();
	float camNearMul = mCamera.getZNear() * 1.001f;

	ID3DXEffect& effect = *CEffectBundle::getInstance().getResourceById( mEffectId );
	
	SVertexXyzDiffuseTex1* vb = NULL;

	// lock buffer
	int lineVertices = mLines.size()*4;
	if( mBufferPosition == 0 || mBufferPosition+lineVertices >= BUFFERS_CAPACITY*4 ) {
		mVB->Lock( 0, 0, (BYTE**)&vb, D3DLOCK_DISCARD );
		mBufferPosition = 0;
	} else {
		mVB->Lock( mBufferPosition*sizeof(SVertexXyzDiffuseTex1), lineVertices*sizeof(SVertexXyzDiffuseTex1), (BYTE**)&vb, D3DLOCK_NOOVERWRITE );
	}

	// fill with lines
	TLinePool::const_iterator it;
	TLinePool::const_iterator itEnd = mLines.end();
	for( it = mLines.begin(); it != itEnd; ++it ) {
		const SLine& line = *it;

		// interpolate
		float alpha = frameTime.getAlpha();
		D3DXVECTOR3 endA, endB;
		D3DXVec3Lerp( &endA, &line.oldInfo.endA, &line.info.endA, alpha );
		D3DXVec3Lerp( &endB, &line.oldInfo.endB, &line.info.endB, alpha );
		float halfWidth = line.oldInfo.halfWidth + alpha * (line.info.halfWidth-line.oldInfo.halfWidth);
		D3DXCOLOR xcolor;
		D3DXColorLerp( &xcolor, &line.oldInfo.color, &line.info.color, alpha );
		D3DCOLOR color = xcolor;

		// line direction on XZ plane
		D3DXVECTOR2 lineDir( endB.x-endA.x, endB.y-endA.y );
		D3DXVec2Normalize( &lineDir, &lineDir );

		// perpendicular to line direction vector, scaled by halfWidth
		float width2 = -lineDir.y * halfWidth;
		float height2 = lineDir.x * halfWidth;

		// transform by camera
		D3DXVECTOR3 corner1( width2, height2,  0.0f );
		D3DXVECTOR3 corner2( -width2, -height2,  0.0f );

		D3DXVECTOR3 c1, c2;
		D3DXVec3TransformCoord( &c1, &corner1, &viewMatrix );
		D3DXVec3TransformCoord( &c2, &corner2, &viewMatrix );
		
		// 4 line corners
		vb->p = endA + c1;
		vb->diffuse = color;
		vb->tu = 0.0f; vb->tv = 0.0f;
		++vb;
		vb->p = endB + c1;
		vb->diffuse = color;
		vb->tu = 1.0f; vb->tv = 0.0f;
		++vb;
		vb->p = endB + c2;
		vb->diffuse = color;
		vb->tu = 1.0f; vb->tv = 1.0f;
		++vb;
		vb->p = endA + c2;
		vb->diffuse = color;
		vb->tu = 0.0f; vb->tv = 1.0f;
		++vb;
	}
	
	// unlock
	mVB->Unlock();

	// draw
	IDirect3DTexture8& texture = *CTextureBundle::getInstance().getResourceById( mTextureId );
	effect.SetTexture( "texBase", &texture );
	UINT passes;
	effect.Begin( &passes, 0 );

	getDevice().SetIndices( mIB, mBufferPosition );
	getDevice().SetStreamSource( 0, mVB, sizeof(SVertexXyzDiffuseTex1) );
	for( UINT p = 0; p < passes; ++p ) {
		effect.Pass( p );
		getDevice().DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, lineVertices, 0, lineVertices/2 );
	}
	effect.End();

	mBufferPosition += lineVertices;
}
